package bcn;

import com.cyberbotics.webots.controller.Robot;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Random;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;


public class Scheduler {
    
    private class BehaviorTask implements Runnable {
        private final Behavior behavior;
        private CountDownLatch latch;
        
        public BehaviorTask(Behavior behavior) {
            this.behavior = behavior;
        }
        
        public void setLatch(CountDownLatch latch) {
        	this.latch = latch;
        }
        
        @Override
        public void run() {
            behavior.step();
            latch.countDown();
        }
    }
    
    
    private final ArrayList<Behavior> behaviors;
    private final ArrayList<BehaviorTask> tasks;
    private final Robot robot;
    private final int controlStep;
    private final Random rnd;
    
    
    public Scheduler(Robot robot, int controlStep, Random rnd) {
        this.behaviors = new ArrayList<Behavior>();
        this.tasks = new ArrayList<BehaviorTask>();
        this.robot = robot;
        this.controlStep = controlStep;
        this.rnd = rnd;
    }
    
    
    public Scheduler(Robot robot, int controlStep, Random rnd, Behavior... behaviors) {
        this(robot, controlStep, rnd);
        add(behaviors);
    }
    
    
    public synchronized void add(Behavior behavior) {
        this.behaviors.add(behavior);
        this.tasks.add(new BehaviorTask(behavior));
        behavior.init();
    }
    
    
    public synchronized void add(Behavior... behaviors) {
        this.behaviors.addAll(Arrays.asList(behaviors));
        for (Behavior b: behaviors) {
            this.tasks.add(new BehaviorTask(b));
            b.init();
        }
    }
    
    
    public synchronized void remove(Behavior behavior) {
    	int id = this.behaviors.indexOf(behavior);
    	if (id >= 0) {
	        this.behaviors.remove(id);
	        this.tasks.remove(id);
        }
    }
    
    
    public void run() {
        ExecutorService pool = Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors());
        while (robot.step(controlStep) != -1) {
            // scheduling behaviors
            synchronized (this) {
            	// shuffle behaviors to simulate parallelism
                Collections.shuffle(tasks, rnd);
                
                /*// sequential execution
                for (BehaviorTask t: tasks) {
                    t.run();
                }*/
                
                // concurrent execution
                CountDownLatch latch = new CountDownLatch(tasks.size());
                for (BehaviorTask t: tasks) {
                	t.setLatch(latch);
                    pool.execute(t);
                }
                // await behaviors termination
                try {
                	latch.await();
                } catch (InterruptedException ex) {
                	ex.printStackTrace();
                }
            }
        }
    }
    
}
